"""
Models to send requests to Zafira server, will be expand as new client features will occur.
They're using camel-case to send as JSON in request directly
"""
import enum


class TestStatus(enum.Enum):
    UNKNOWN = 'UNKNOWN'
    IN_PROGRESS = 'IN_PROGRESS'
    PASSED = 'PASSED'
    FAILED = 'FAILED'
    SKIPPED = 'SKIPPED'
    ABORTED = 'ABORTED'
    QUEUED = 'QUEUED'


class Initiator(enum.Enum):
    SCHEDULER = 'SCHEDULER'
    UPSTREAM_JOB = 'UPSTREAM_JOB'
    HUMAN = 'HUMAN'


class DriverMode(enum.Enum):
    METHOD_MODE = "METHOD_MODE"
    CLASS_MODE = "CLASS_MODE"
    SUITE_MODE = "SUITE_MODE"



class ZafiraError(Exception):
    """ Common exception for Zafira App """


class ConfigError(ZafiraError):
    """ Raises if environ var is missing """


class APIError(ZafiraError):
    """ An exception for Zafira client API calls issues """



__all__ = ['zafira_state', 'client', 'resource_constants', 'resources', 'plugin', 'listener']

from .zafira_state import ZafiraState
from .client import ZafiraClient
from .plugin import connector_obj
from .listener import PyTestZafiraListener



import configparser
import logging
from os import getcwd, environ
from enum import Enum

from .exceptions import ConfigError

logger = logging.getLogger('settings')

CONFIG_FILE_PATH = getcwd() + '/zafira_properties.ini'


class Context:

    @staticmethod
    def get(parameter):
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE_PATH)
        return config.get('config', parameter.value)


class Parameter(Enum):
    SERVICE_URL = 'service-url'
    ACCESS_TOKEN = 'access_token'
    ZAFIRA_ENABLED = 'zafira_enabled'
    JOB_NAME = 'job_name'
    ARTIFACT_EXPIRES_IN_DEFAULT_TIME = 'artifact_expires_in_default_time'
    ARTIFACT_LOG_NAME = 'artifact_log_name'
    AWS_SCREEN_SHOT_BUCKET = 'aws_screen_shot_bucket'
    S3_SAVE_SCREENSHOTS = 's3_save_screenshots'


def get_env_var(env_var_key):
    """
    Getter for environment variable keys.  Uses `os.environ.get(KEYNAME)`
    :param env_key:  Type string.  Key name of environment variable to get.
    :return:  Value of the environment variable.
    """
    env_var_value = environ.get(env_var_key)
    if env_var_value is None:
        logger.error('ENV var missing: [{0}], please set this variable'.format(env_var_key))
        raise ConfigError("Env variable {0} is mandatory, please set this variable".format(env_var_key))
    else:
        logger.debug('ENV variable is: [{0}]:[{1}]'.format(env_var_key, env_var_value))
    return env_var_value



from .resource_constants import Initiator, TestStatus, DriverMode
from .api_request import APIRequest
from .resources import user_cred
from .resources import test_suite
from .resources import test_case
from .resources import job
from .resources import test_run
from .resources import test
from .resources import artifact
from .resources import user
from .resources import refresh_token


class ZafiraClient:
    DEFAULT_USER = "anonymous"
    STATUS_PATH = "/api/status"
    REFRESH_TOKEN_PATH = "/api/iam/v1/auth/refresh"
    LOGIN_PATH = "/api/auth/login"
    PROFILE_PATH = "/api/users/profile"
    TEST_SUITES_PATH = "/api/tests/suites"
    TEST_CASES_PATH = "/api/tests/cases"
    JOBS_PATH = "/api/jobs"
    TEST_RUNS_PATH = "/api/tests/runs"
    TESTS_PATH = "/api/tests"
    TEST_FINISH_PATH = "/api/tests/{}/finish"
    TEST_RUNS_FINISH_PATH = "/api/tests/runs/{}/finish"
    TEST_RUNS_ABORT_PATH = "/api/tests/runs/abort?id={}"
    TEST_BY_ID_PATH = "/api/tests/{}"
    ADD_TEST_ARTIFACT = "/api/tests/{}/artifacts"
    TEST_WORK_ITEMS_PATH =  "/api/tests/{}/workitems"
    TEST_RUNS_RESULTS_PATH = "/api/tests/runs/{}/results"
    USERS_PATH = "/api/users"
    PROJECT_PATH = "/api/projects"
    GET_SETTING_TOOL_PATH = "/api/settings/tool/{}"

    INSTANCE = None

    def __new__(cls, service_url):
        if not cls.INSTANCE:
            cls.INSTANCE = super(ZafiraClient, cls).__new__(cls)
        return cls.INSTANCE

    def __init__(self, service_url):
        self.api = APIRequest(service_url)
        self.auth_token = ''

    def get_setting_tool(self, tool):
        return self.api.send_get(ZafiraClient.GET_SETTING_TOOL_PATH.format(tool),
                                 headers=self.init_auth_headers(), default_err_msg="Unable to get settings by tool")

    def is_zafira_available(self):
        is_available = False
        if self.api.send_get(ZafiraClient.STATUS_PATH, default_err_msg="Unable to send ping").status_code == 200:
            is_available = True
        return is_available

    def login(self, username, password):
        user_cred['username'] = username
        user_cred['password'] = password
        return self.api.send_post_without_authorization(ZafiraClient.LOGIN_PATH, user_cred, 'Unable to login')

    def create_test_suite(self, user_id, suite_name, file_name, description=None):
        test_suite['userId'] = user_id
        test_suite['description'] = description
        test_suite['fileName'] = file_name
        test_suite['name'] = suite_name
        return self.api.send_post(ZafiraClient.TEST_SUITES_PATH, test_suite, headers=self.init_auth_headers(),
                                  default_err_msg="Unable to create test suite")

    def create_test_case(self, test_class, test_method, test_suite_id, user_id, info=None, project=None):
        test_case['testClass'] = test_class
        test_case['testMethod'] = test_method
        test_case['testSuiteId'] = test_suite_id
        test_case['primaryOwnerId'] = user_id
        test_case['info'] = info
        test_case['project'] = project
        return self.api.send_post(ZafiraClient.TEST_CASES_PATH, test_case, headers=self.init_headers_with_project(),
                                  default_err_msg="Unable to create test case")

    def create_job(self, user_id, job_name, job_url, jenkins_host):
        job["userId"] = user_id
        job["jobURL"] = job_url
        job["name"] = job_name
        job["jenkinsHost"] = jenkins_host
        return self.api.send_post(ZafiraClient.JOBS_PATH, job, headers=self.init_auth_headers(),
                                  default_err_msg="Unable to create job")

    def start_test_run(self, job_id, test_suite_id, build_number, started_by=Initiator.SCHEDULER.value,  # noqa F405
                       driver_mode=DriverMode.METHOD_MODE.value, config=None, blocker=None,  # noqa F405
                       work_item=None, status=None, project=None, known_issue=None):
        test_run["jobId"] = job_id
        test_run["testSuiteId"] = test_suite_id
        test_run["buildNumber"] = build_number
        test_run["startedBy"] = started_by
        test_run["driverMode"] = driver_mode
        test_run["blocker"] = blocker
        test_run["workItem"] = work_item
        test_run["status"] = status
        test_run["project"] = project
        test_run["knownIssue"] = known_issue
        test_run["configXML"] = config
        return self.api.send_post(ZafiraClient.TEST_RUNS_PATH, test_run, headers=self.init_auth_headers(),
                                  default_err_msg="Unable to start test run")

    def update_test_run(self, test_run):
        return self.api.send_put(ZafiraClient.TEST_RUNS_PATH, test_run, headers=self.init_auth_headers(),
                                 default_err_msg="Unable to start test run")

    def finish_test_run(self, test_run_id):
        return self.api.send_post(ZafiraClient.TEST_RUNS_FINISH_PATH.format(test_run_id),
                                  headers=self.init_auth_headers(), default_err_msg="Unable to finish test run")

    def abort_test_run(self, test_run_id):
        return self.api.send_post(ZafiraClient.TEST_RUNS_ABORT_PATH.format(test_run_id),
                                  headers=self.init_auth_headers(), default_err_msg="Unable to find test run by id")

    def start_test(self, test_run_id, test_case_id, test_name, start_time, ci_test_id,
                   status=TestStatus.IN_PROGRESS.value, test_class=None,  # noqa F405
                   test_group=None, work_items=None):
        test["testRunId"] = test_run_id
        test["testCaseId"] = test_case_id
        test["name"] = test_name
        test["ciTestId"] = ci_test_id
        test["startTime"] = start_time
        test["testClass"] = test_class
        test["status"] = status
        test["workItems"] = work_items
        test["testGroup"] = test_group
        return self.api.send_post(ZafiraClient.TESTS_PATH, test, headers=self.init_auth_headers(),
                                  default_err_msg="Unable to start test")

    def finish_test(self, test):
        return self.api.send_post(ZafiraClient.TEST_FINISH_PATH.format(test["id"]), test,
                                  headers=self.init_auth_headers(), default_err_msg="Unable to finish test")

    def delete_test_by_id(self, test_id):
        return self.api.send_delete(ZafiraClient.TEST_BY_ID_PATH.format(test_id), headers=self.init_auth_headers(),
                                    default_err_msg="Unable to finish test")

    def update_test(self, test):
        return self.api.send_put(ZafiraClient.TESTS_PATH, test, headers=self.init_auth_headers(),
                                 default_err_msg="Unable to update test")

    def add_test_artifact_to_test(self, test_id, link, artifact_name, expires_in=None):
        artifact["testId"] = test_id
        artifact["link"] = link
        artifact["name"] = artifact_name
        artifact["expiresIn"] = expires_in
        return self.api.send_post(ZafiraClient.ADD_TEST_ARTIFACT.format(test_id), artifact,
                                  headers=self.init_auth_headers(), default_err_msg="Unable to add test artifact")

    def refresh_token(self, token):
        refresh_token["refreshToken"] = token
        return self.api.send_post_without_authorization(ZafiraClient.REFRESH_TOKEN_PATH, refresh_token,
                                                        default_err_msg="Unable to refresh token")

    def create_test_work_items(self, test_id, list_work_items):
        return self.api.send_post(ZafiraClient.TEST_WORK_ITEMS_PATH.format(test_id), list_work_items,
                                  headers=self.init_auth_headers(), default_err_msg="Unable to create workitems")

    def get_test_run_results(self, test_run_id):
        return self.api.send_get(ZafiraClient.TEST_RUNS_RESULTS_PATH.format(test_run_id),
                                 headers=self.init_auth_headers(), default_err_msg="Unable to find test run results")

    def create_user(self, first_name, last_name, email, password):
        user["firstName"] = first_name
        user["lastName"] = last_name
        user["email"] = email
        user["password"] = password
        return self.api.send_put(ZafiraClient.USERS_PATH, user, headers=self.init_auth_headers(),
                                 default_err_msg="Unable to create user")

    def get_user_profile(self, username=None):
        endpoint = ZafiraClient.PROFILE_PATH
        if username:
            endpoint += "?username=" + username
        return self.api.send_get(endpoint, headers=self.init_auth_headers(),
                                 default_err_msg="Unable to authorize user")

    def get_user_or_anonymous_if_not_found(self, username):
        try:
            return self.get_user_profile(username)
        except Exception:
            return self.get_user_profile(self.DEFAULT_USER)

    def create_project(self, project):
        return self.api.send_post(ZafiraClient.PROJECT_PATH, project, headers=self.init_auth_headers(),
                                  default_err_msg="Unable to create project")

    def init_auth_headers(self):
        return {"Authorization": "Bearer " + self.auth_token}

    def init_headers_with_project(self):
        return {"Authorization": "Bearer " + self.auth_token, "project": "UNKNOWN"}



job = {
  "id": None,
  "jenkinsHost": "http://127.0.0.1:8080",
  "jobURL": "http://localhost:8080/job/Job_Selenoid/",
  "name": "NAME",
  "userId": "USER_ID_INT"
}


project = {
    "description": "DESCRIPTION",
    "id": None,
    "name": "NAME"
}


refresh_token = {
  "refreshToken": "REFRESH_TOKEN"
}


artifact = {
      "expiresIn": None,
      "id": None,
      "link": "LINK",
      "name": "NAME",
      "testId": "TEST_ID_INT"
}


test_case = {
  "id": None,
  "info": None,
  "primaryOwnerId": "OWNER_ID_INT",
  "project": None,
  "secondaryOwnerId": None,
  "testClass": "TEST_CLASS",
  "testMethod": "TEST_METHOD",
  "testSuiteId": "TEST_SUITE_ID_INT"
}


test = {
  "artifacts": [],
  "blocker": None,
  "ciTestId": None,
  "configXML": None,
  "dependsOnMethods": None,
  "finishTime": None,
  "id": None,
  "knownIssue": None,
  "message": None,
  "messageHashCode": None,
  "name": "NAME",
  "needRerun": None,
  "retry": None,
  "startTime": None,
  "status": None,
  "testArgs": None,
  "testCaseId": "TEST_CASE_ID_INT",
  "testClass": None,
  "testGroup": None,
  "testMetrics": None,
  "testRunId": "TEST_RUN_ID_INT",
  "workItems": []
}


test_run = {
  "blocker": None,
  "buildNumber": "BUILD_NUMBER_INT",
  "ciRunId": None,
  "configXML": None,
  "driverMode": "DRIVER_MODE",
  "id": None,
  "jobId": "JOB_ID_INT",
  "knownIssue": None,
  "project": None,
  "reviewed": None,
  "scmBranch": None,
  "scmCommit": None,
  "scmURL": None,
  "startedBy": "STARTED_BY",
  "status": None,
  "testSuiteId": "TEST_SUITE_ID_INT",
  "upstreamJobBuildNumber": None,
  "upstreamJobId": None,
  "userId": None,
  "workItem": None
}


test_suite = {
  "description": None,
  "fileName": "FILE_NAME",
  "name": "NAME",
  "userId": "USER_ID"
}


user_cred = {
  "password": "PASSWORD",
  "username": "USERNAME"
}


user = {
  "email": None,
  "firstName": "FIRST_NAME",
  "id": None,
  "lastLogin": None,
  "lastName": "LAST_NAME",
  "password": "PASSWORD",
  "permissions": [],
  "photoURL": None,
  "preferences": [],
  "roles": [],
  "source": None,
  "status": None,
  "username": "USERNAME"
}



from distutils.core import setup
setup(
  name = 'zebrunnerpy',
  packages = ['zebrunnerpy'],
  version = '0.1.3',
  license='MIT',
  description = 'Python3 connector for Zebrunner-reporting',
  author = 'Sergey Shukalovich',
  author_email = 'sshukalovich@solvd.com',
  url = 'https://github.com/shukal94/zebrunner_py',
  download_url = 'https://github.com/shukal94/zebrunner_py',
  keywords = ['automation', 'zebrunner', 'testing'],
  install_requires=[
          'pytest',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)



import logging


class BaseZafiraListener:
    logger = logging.getLogger('zafira')

    def __init__(self, state):
        self.state = state

        self.initialized = self.initialize_zafira()

    def initialize_zafira(self):
        enabled = False
        try:
            if self.state.is_enabled:

                self.state.is_enabled = self.state.zc.is_zafira_available()

                if self.state.is_enabled:
                    self.state.refresh_token = self.state.zc.refresh_token(self.state.access_token).json()
                    self.state.zc.auth_token = self.state.refresh_token['authToken']
                    self.logger.info("Zafira is " + "available" if self.state.is_enabled else "unavailable")

            enabled = self.state.is_enabled
        except Exception as e:
            self.logger.error("Unable to initialize connector: {}".format(e))
        return enabled

    def compose_package_name(self, path_entries_list):
        if len(path_entries_list) == 2:
            return path_entries_list[0]
        return path_entries_list[0] + '/' + self.compose_package_name(path_entries_list[1:])

    def add_artifact_to_test(self, test, artifact_name, artifact_link, expires_in=None):
        """
        Adds test artifact to test
        """
        try:
            self.state.zc.add_test_artifact_to_test(test["id"], artifact_link, artifact_name, expires_in)
        except Exception as e:
            self.logger.error("Unable to add artifact to test correctly: {}".format(e))

    def on_test_success(self):
        """
        An abstract handler for posting successfull test onto Zafira UI
        """

    def on_test_failure(self, message):
        """
        An abstract handler for posting failed test onto Zafira UI
        :param message: error message from failing test
        """

    def on_test_skipped(self, message):
        """
        An abstract handler for posting skipped test onto Zafira UI
        :param message: skip reason
        """

    def add_work_item_to_test(self, test_id, work_item):
        if not self.state.is_enabled:
            return
        try:
            work_items = list()
            work_items.append(work_item if len(work_item) < self.state.MAX_LENGTH_OF_WORKITEM else 'Skipped')
            self.state.zc.create_test_work_items(test_id, work_items)
        except Exception as e:
            self.logger.error("Unable to add work item: {}".format(e))


import time
import pytest
import uuid

from .context import Context, Parameter
from .base_listener import BaseZafiraListener
from .resource_constants import TestStatus


class PyTestZafiraListener(BaseZafiraListener):
    """
    Contains hook implementations
    """

    def __init__(self, state):
        super().__init__(state)

        self.state.CONFIG = "<config><arg unique='false'><key>env</key><value>API " \
                            + Context.get_gui_parameter(Parameter.BROWSER).upper() + "</value></arg></config>"

    @pytest.hookimpl
    def pytest_sessionstart(self, session):
        """
        Setup-class handler, signs in user, creates a testsuite,
        testcase, job and registers testrun in Zafira
        """
        if not self.initialized:
            return
        try:
            self.state.user = self.state.zc.get_user_profile().json()
            self.state.test_suite = self.state.zc.create_test_suite(
                self.state.user["id"], self.state.suite_name, 'filename'
            ).json()
            self.state.job = self.state.zc.create_job(
                self.state.user["id"],
                self.state.job_name,
                self.state.job_url,
                "jenkins_host"
            ).json()
            self.state.test_run = self.state.zc.start_test_run(
                self.state.job["id"],
                self.state.test_suite["id"],
                0,
                config=self.state.CONFIG
            ).json()
            self.state.ci_run_id = self.state.test_run.get('ciRunId')
        except Exception as e:
            self.state.is_enabled = False
            self.logger.error("Undefined error during test run registration! {}".format(e))

    @pytest.hookimpl
    def pytest_runtest_setup(self, item):
        """
        Setup handler, set up initial parameters for test,
        attaches to testsuite, registers and starts the test
        """
        if not self.state.is_enabled:
            return
        try:
            test_name = item.name
            class_name = item.nodeid.split('::')[1]
            self.state.ci_test_id = str(uuid.uuid4())

            full_path_to_file = item.nodeid.split('::')[0].split('/')
            package = self.compose_package_name(full_path_to_file) + '/'
            self.state.test_case = self.state.zc.create_test_case(
                class_name, test_name, self.state.test_suite["id"], self.state.user["id"]
            ).json()
            work_items = []
            if hasattr(item._evalxfail, 'reason'):
                work_items.append('xfail')
            self.state.test = self.state.zc.start_test(
                self.state.test_run["id"],
                self.state.test_case["id"],
                test_name,
                round(time.time() * 1000),
                self.state.ci_test_id,
                TestStatus.IN_PROGRESS.value,
                class_name,
                package,
                work_items
            ).json()
        except Exception as e:
            self.logger.error("Undefined error during test case/method start! {}".format(e))

    @pytest.hookimpl
    def pytest_runtest_teardown(self, item):
        """
        Teardown handler. Finishes test, adds workitems if needed
        """
        if not self.state.is_enabled:
            return
        try:
            if item._skipped_by_mark:
                test_name = item.name
                class_name = item.nodeid.split('::')[1]
                full_path_to_file = item.nodeid.split('::')[0].split('/')
                package = self.compose_package_name(full_path_to_file) + '/'
                self.state.test_case = self.state.zc.create_test_case(
                    class_name, test_name, self.state.test_suite["id"], self.state.user["id"]
                ).json()
                self.state.test = self.state.zc.start_test(
                    self.state.test_run["id"], self.state.test_case["id"], test_name,
                    round(time.time() * 1000), self.state.ci_test_id, test_class=class_name, test_group=package
                ).json()

                self.state.test['status'] = TestStatus.SKIPPED.value
                self.add_work_item_to_test(self.state.test['id'], self.state.skip_reason)

            self.state.zc.finish_test(self.state.test)
        except Exception as e:
            self.logger.error("Unable to finish test run correctly: {}".format(e))

    @pytest.hookimpl
    def pytest_runtest_logreport(self, report):
        """
        Set test status, stacktrace if needed
        :param report: info about test
        """
        if not self.state.is_enabled:
            return
        try:
            if report.when == 'setup':
                if report.skipped:
                    self.state.skip_reason = report.longrepr[2]
                if report.failed:
                    self.on_test_failure(report)
            if report.when == 'call':
                self.state.test["finishTime"] = round(time.time() * 1000)
                test_result = report.outcome
                if test_result is 'passed':
                    self.on_test_success()
                elif test_result is 'failed':
                    self.on_test_failure(report)
                else:
                    self.on_test_skipped(report)
                self.add_artifact_to_test(
                    self.state.test,
                    self.state.artifact_log_name,
                    'http://google.com',
                    self.state.artifact_expires_in_default_time
                )
        except Exception as e:
            self.logger.error("Unable to finish test correctly: {}".format(e))

    @pytest.hookimpl
    def pytest_sessionfinish(self, session, exitstatus):
        """
        Teardown-class handler, closes the testrun
        """
        if not self.state.is_enabled:
            return
        try:
            self.state.zc.finish_test_run(self.state.test_run["id"])
        except Exception as e:
            self.logger.error("Unable to finish test run correctly: {}".format(e))

    def on_test_success(self):
        self.state.test['status'] = TestStatus.PASSED.value

    def on_test_failure(self, message):
        self.state.test['status'] = TestStatus.FAILED.value
        self.state.test['message'] = message.longreprtext

    def on_test_skipped(self, message):
        self.state.test['message'] = message.longreprtext
        if not hasattr(message, 'wasxfail'):
            self.state.test['status'] = TestStatus.SKIPPED.value
        else:
            self.state.test['status'] = TestStatus.FAILED.value
